module Chess where

import Color (..)
import Dict
import List
import Maybe
import Mouse
import Signal (..)
import Graphics.Collage (..)
import Graphics.Element (..)
import Text (..)
import Window (..)

import Model.Board    as Board
import Model.Game     (..)
import Model.Game     as Game
import Model.Piece    (..)
import Model.Player   (..)
import Model.Position (..)
import Model.Position as Position
import Model.Move     (..)
import Model.Move     as Move

type Action = MouseMove (Int, Int) | MouseClick

updates : Signal Action
updates = mergeMany [ map MouseMove Mouse.position
                    , map (always MouseClick) Mouse.clicks ]



type alias State = { game          : Game
                   , moveText      : String
                   , mousePosition : (Int, Int)
                   , selectedField : Maybe Position.Position
                   , possibleMoves : List Move
                   , size          : Int }

initial : State
initial = { game          = Game.initial
          , moveText      = ""
          , mousePosition = (0, 0)
          , selectedField = Nothing
          , possibleMoves = []
          , size          = 80 }

select : State -> State
select s =
    let size    = s.size
        board   = s.game.board
        player  = s.game.player
        (x, y)  = s.mousePosition
        f       = x // size + 1
        r       = 8 - y // size
        isValid = let piece = Dict.get (f, r) board
                  in  piece `Maybe.andThen` (\p -> Just <| p `belongsTo` player) == Just True
        filter m =
            case m of
                Normal m    -> m.to == (f, r)
                EnPassant m -> m.to == (f, r)
                Promotion m -> m.to == (f, r)
                KingsideCastling m -> m.to == (f, r)
                QueensideCastling m -> m.to == (f, r)
        mg      = List.filter filter s.possibleMoves
    in case mg of
        [] -> if isValid
            then { s | selectedField <- Just (f, r)
                     , possibleMoves <- legalMoves s.game (f, r) }
            else { s | selectedField <- Nothing
                     , possibleMoves <- [] }
        [m] -> { s | game <- applyMove s.game m
                       , selectedField <- Nothing
                       , possibleMoves <- []
                       , moveText      <- let movenumber = if s.game.player == White then toString s.game.fullmoves ++ ". " else ""
                                          in s.moveText ++ " " ++ movenumber ++ Move.show s.game m }
        [k, b, r, q] -> { s | game <- applyMove s.game q
                       , selectedField <- Nothing
                       , possibleMoves <- []
                       , moveText      <- let movenumber = if s.game.player == White then toString s.game.fullmoves ++ ". " else ""
                                          in s.moveText ++ " " ++ movenumber ++ Move.show s.game q } -- TODO pawn promotion

step : Action -> State -> State
step a s =
    case a of
        MouseMove (x, y) -> { s | mousePosition <- (x, y) }
        MouseClick       -> select s

-- View
board : State -> List Form
board s =
    let board             = s.game.board
        selected          = s.selectedField
        possibleMoves     = s.possibleMoves
        size              = toFloat s.size
        lightField        = filled grey      <| rect size size
        darkField         = filled charcoal  <| rect size size
        selectedField     = filled blue      <| rect size size
        lightMoveField    = filled green     <| rect size size
        darkMoveField     = filled darkGreen <| rect size size
        lightCaptureField = filled red       <| rect size size
        darkCaptureField  = filled darkRed   <| rect size size
        translate r f = move (size * (-4.5 + toFloat r), size * (-4.5 + toFloat f))
        mapper m =
            case m of
                Normal m    -> m.to
                EnPassant m -> m.to
                Promotion m -> m.to
                KingsideCastling m -> m.to
                QueensideCastling m -> m.to
        display = \(r, f) -> translate r f <|
            if selected `Maybe.andThen` (\(sr, sf) -> Just (r == sr && f == sf)) == Just True
                then selectedField
                else if (r, f) `List.member` (List.map mapper possibleMoves)
                    then if (r, f) `Dict.member` board
                        then if (r + f) % 2 == 0
                            then darkCaptureField
                            else lightCaptureField
                        else if (r + f) % 2 == 0
                            then darkMoveField
                            else lightMoveField
                    else if (r + f) % 2 == 0
                        then darkField
                        else lightField
    in  List.map display Position.all

pieces : State -> List Form
pieces s =
    let board       = s.game.board
        size        = toFloat s.size
        translate   = \(r, f) -> move (size * (toFloat r - 4.5), size * (toFloat f - 4.5))
        display k v = translate k << toForm << image 60 60 <| "img/" ++ toString v ++ ".gif"
    in Dict.values <| Dict.map display board

chessboard : State -> Element
chessboard s = collage 640 640 <| board s ++ pieces s

scene : State -> Element
scene s = flow right [chessboard s]

main : Signal Element
main = map scene <| foldp step initial updates
