module Model.Move where

import Dict
import List

import Util (..)

import Model.Board    (..)
import Model.Board    as Board
import Model.Game     (..)
import Model.Piece    (..)
import Model.Piece    as Piece
import Model.Player   (..)
import Model.Position (..)
import Model.Position as Position


-- Helper ----------------------------------------------------------------------

isFree : Game -> Position -> Bool
isFree g p = Position.isValid p && (not <| Dict.member p g.board)

isOpposite : Game -> Position -> Bool
isOpposite g p =
    let piece = Dict.get p g.board
    in case piece of
        Nothing -> False
        Just p  -> not <| p `belongsTo` g.player


-- Move constructors -----------------------------------------------------------

normalNewBoard : Position -> Position -> Board -> Board
normalNewBoard from to b =
    let p = Dict.get from b
    in b |> Dict.remove from |> Dict.update to (always p)

normalNewCastling : Position -> Castling -> Castling
normalNewCastling from c =
    { c | whiteKingside  <- if from == e1 || from == h1 then False else c.whiteKingside
        , whiteQueenside <- if from == e1 || from == a1 then False else c.whiteQueenside
        , blackKingside  <- if from == e8 || from == h8 then False else c.blackKingside
        , blackQueenside <- if from == e8 || from == a8 then False else c.blackQueenside }

normalNewEnPassant : Position -> Position -> Board -> Maybe Position
normalNewEnPassant = \(ff, fr) (tf, tr) b ->
    let (Just p) = Dict.get (ff, fr) b
    in case p of
        WhitePawn -> if tr - fr == 2 then Just (tf, tr - 1) else Nothing
        BlackPawn -> if fr - tr == 2 then Just (tf, tr + 1) else Nothing
        _         -> Nothing

normalNewHalfmoves : Position -> Position -> Board -> Int -> Int
normalNewHalfmoves from to b h =
    let (Just fp) = Dict.get from b
        tp        = Dict.get to   b
    in case fp of
        WhitePawn -> 0
        BlackPawn -> 0
        _         -> case tp of
            Nothing -> h + 1
            _       -> 0

normal : Position -> Position -> Game -> Game
normal from to g =
    { g | board     <- normalNewBoard from to g.board
        , player    <- opponent g.player
        , castling  <- normalNewCastling from g.castling
        , enPassant <- normalNewEnPassant from to g.board
        , halfmoves <- normalNewHalfmoves from to g.board g.halfmoves
        , fullmoves <- if g.player == White then g.fullmoves else g.fullmoves + 1 }

enPassantNewBoard : Position -> Position -> Board -> Board
enPassantNewBoard = \(ff, fr) (tf, tr) b ->
    let p = Dict.get (ff, fr) b
    in b |> Dict.remove (ff, fr) |> Dict.update (tf, tr) (always p) |> Dict.remove (tf, fr)

enPassant : Position -> Position -> Game -> Game
enPassant from to g =
    { g | board     <- enPassantNewBoard from to g.board
        , player    <- opponent g.player
        , enPassant <- Nothing
        , halfmoves <- 0
        , fullmoves <- if g.player == White then g.fullmoves else g.fullmoves + 1 }

promotionNewBoard : Position -> Position -> Piece -> Board -> Board
promotionNewBoard from to p b =
    b |> Dict.remove from |> Dict.update to (always <| Just p)

promotion : Position -> Position -> Piece -> Game -> Game
promotion from to piece g =
    let isWhite = g.player == White
    in { g | board     <- promotionNewBoard from to piece g.board
           , player    <- opponent g.player
           , enPassant <- Nothing
           , halfmoves <- 0
           , fullmoves <- if g.player == White then g.fullmoves else g.fullmoves + 1 } -- Problem: Promotion moves have the same key --> key should not be to position, just store list of next states

kingsideCastlingNewBoard : Player -> Board -> Board
kingsideCastlingNewBoard p b =
    let isWhite = p == White
        rank    = if isWhite then 1 else 8
        king    = if isWhite then WhiteKing else BlackKing
        rook    = if isWhite then WhiteRook else BlackRook
    in b |> Dict.remove (Position.e, rank)
         |> Dict.update (g         , rank) (always <| Just king)
         |> Dict.remove (h         , rank)
         |> Dict.update (f         , rank) (always <| Just rook)

castlingNewCastling : Player -> Castling -> Castling
castlingNewCastling p c =
    let isWhite = p == White
    in { c | whiteKingside  <- if isWhite then False else c.whiteKingside
           , whiteQueenside <- if isWhite then False else c.whiteQueenside
           , blackKingside  <- if isWhite then c.blackKingside else False
           , blackQueenside <- if isWhite then c.blackQueenside else False }

kingsideCastling : Game -> Game
kingsideCastling g =
    let isWhite = g.player == White
    in { g | board     <- kingsideCastlingNewBoard g.player g.board
           , player    <- opponent g.player
           , castling  <- castlingNewCastling g.player g.castling
           , enPassant <- Nothing
           , halfmoves <- g.halfmoves + 1
           , fullmoves <- if isWhite then g.fullmoves else g.fullmoves + 1 }

queensideCastlingNewBoard : Player -> Board -> Board
queensideCastlingNewBoard p b =
    let isWhite = p == White
        rank    = if isWhite then 1 else 8
        king    = if isWhite then WhiteKing else BlackKing
        rook    = if isWhite then WhiteRook else BlackRook
    in b |> Dict.remove (Position.e, rank)
         |> Dict.update (c, rank) (always <| Just king)
         |> Dict.remove (a, rank)
         |> Dict.update (d, rank) (always <| Just rook)

queensideCastling : Game -> Game
queensideCastling g =
    let isWhite = g.player == White
    in { g | board     <- queensideCastlingNewBoard g.player g.board
           , player    <- opponent g.player
           , castling  <- castlingNewCastling g.player g.castling
           , enPassant <- Nothing
           , halfmoves <- g.halfmoves + 1
           , fullmoves <- if isWhite then g.fullmoves else g.fullmoves + 1 }


-- Captures --------------------------------------------------------------------

generateWhitePawnCaptures : Game -> Position -> List Position
generateWhitePawnCaptures g = \(f, r) ->
    let potential = [ (f - 1, r + 1),
                      (f + 1, r + 1) ]
    in List.filter (isOpposite g) potential

generateBlackPawnCaptures : Game -> Position -> List Position
generateBlackPawnCaptures g = \(f, r) ->
    let potential = [ (f - 1, r - 1),
                      (f + 1, r - 1) ]
    in List.filter (isOpposite g) potential

generateKnightCaptures : Game -> Position -> List Position
generateKnightCaptures g = \(f, r) ->
    let potential = [ (f - 2, r - 1)
                    , (f - 2, r + 1)
                    , (f - 1, r - 2)
                    , (f - 1, r + 2)
                    , (f + 1, r - 2)
                    , (f + 1, r + 2)
                    , (f + 2, r - 1)
                    , (f + 2, r + 1) ]
    in List.filter (isOpposite g) potential

generateSlidingCaptures : Game -> Position -> (File -> Int -> Int) -> (Rank -> Int -> Int) -> List Position
generateSlidingCaptures g = \(f, r) df dr ->
    let potential = List.map (\i -> (df f i, dr r i)) [1..7]
        blocked   = dropWhile (isFree g) potential
    in case blocked of
            []      -> []
            (x::xs) -> if isOpposite g x then [x] else []

generateBishopCaptures : Game -> Position -> List Position
generateBishopCaptures g p =
    let slide = generateSlidingCaptures g p
    in     slide (-) (-) -- down, left
        ++ slide (-) (+) -- down, right
        ++ slide (+) (-) -- up, left
        ++ slide (+) (+) -- up, right

generateRookCaptures : Game -> Position -> List Position
generateRookCaptures g p =
    let slide = generateSlidingCaptures g p
    in     slide (-)    always -- down
        ++ slide always (-)    -- left
        ++ slide always (+)    -- right
        ++ slide (+)    always -- up

generateQueenCaptures : Game -> Position -> List Position
generateQueenCaptures g p =
    let slide = generateSlidingCaptures g p
    in     slide (-)    (-)    -- down, left
        ++ slide (-)    always -- down
        ++ slide (-)    (+)    -- down, right
        ++ slide always (-)    -- left
        ++ slide always (+)    -- right
        ++ slide (+)    (-)    -- up, left
        ++ slide (+)    always -- up
        ++ slide (+)    (+)    -- up, right

generateKingCaptures : Game -> Position -> List Position
generateKingCaptures g = \(f, r) ->
    let potential = [ (f - 1, r - 1)
                    , (f - 1, r    )
                    , (f - 1, r + 1)
                    , (f    , r - 1)
                    , (f    , r + 1)
                    , (f + 1, r - 1)
                    , (f + 1, r    )
                    , (f + 1, r + 1) ]
    in List.filter (isOpposite g) potential

generateCaptures : Game -> List Position
generateCaptures g =
    let positions = Dict.keys <| Dict.filter (\pos piece -> piece `belongsTo` g.player) g.board
        mapper    = \from -> let (Just p) = Dict.get from g.board
            in case p of
                WhitePawn   -> generateWhitePawnCaptures g from
                BlackPawn   -> generateBlackPawnCaptures g from
                WhiteKnight -> generateKnightCaptures    g from
                BlackKnight -> generateKnightCaptures    g from
                WhiteBishop -> generateBishopCaptures    g from
                BlackBishop -> generateBishopCaptures    g from
                WhiteRook   -> generateRookCaptures      g from
                BlackRook   -> generateRookCaptures      g from
                WhiteQueen  -> generateQueenCaptures     g from
                BlackQueen  -> generateQueenCaptures     g from
                WhiteKing   -> generateKingCaptures      g from
                BlackKing   -> generateKingCaptures      g from
    in List.concatMap mapper positions

isCheckmate : Game -> Bool
isCheckmate g = isCheck { g | player <- opponent g.player } && allLegalMoves g == []

isStalemate : Game -> Bool
isStalemate g = (not <| isCheck { g | player <- opponent g.player }) && allLegalMoves g == []

isCheck : Game -> Bool
isCheck g =
    let captures = generateCaptures g
        isKing   = \(Just p) -> p == WhiteKing || p == BlackKing
    in List.any (\p -> isKing <| Dict.get p g.board) captures

-- Move type -------------------------------------------------------------------

type Move = Normal            { from : Position, to : Position }
          | EnPassant         { from : Position, to : Position }
          | Promotion         { from : Position, to : Position, piece : Piece }
          | KingsideCastling  { to : Position } -- to not neccessary
          | QueensideCastling { to : Position } -- to not neccessary

normal' : Position -> Position -> Move
normal' from to = Normal { from = from, to = to }

enPassant' : Position -> Position -> Move
enPassant' from to = EnPassant { from = from, to = to }

promotion' : Position -> Position -> Piece -> Move
promotion' from to piece = Promotion { from = from, to = to, piece = piece }


potentialWhitePawnNormalMoves : Game -> Position -> List Move
potentialWhitePawnNormalMoves g = \(f, r) ->
    if r == 7
        then []
        else
            let left =
                    if isOpposite g (f - 1, r + 1)
                        then [normal' (f, r) (f - 1, r + 1)]
                        else []
                forward =
                    if isFree g (f, r + 1)
                        then [normal' (f, r) (f, r + 1)]
                        else []
                fastForward =
                    if r == 2 && isFree g (f, 3) && isFree g (f, 4)
                        then [normal' (f, 2) (f, 4)]
                        else []
                right =
                    if isOpposite g (f + 1, r + 1)
                        then [normal' (f, r) (f + 1, r + 1)]
                        else []
            in  left ++ forward ++ fastForward ++ right

potentialWhiteEnPassant : Game -> Position -> List Move
potentialWhiteEnPassant g = \(f, r) ->
    case g.enPassant of
        Nothing         -> []
        Just (epf, epr) ->
            if r + 1 == epr && abs (f - epf) == 1
                then [enPassant' (f, r) (epf, epr)]
                else []

potentialWhitePromotions : Game -> Position -> List Move
potentialWhitePromotions g = \(f, r) ->
    if r == 7
        then
            let options = [WhiteKnight, WhiteBishop, WhiteRook, WhiteQueen]
                left =
                    if isOpposite g (f - 1, 8)
                        then List.map (promotion' (f, r) (f - 1, 8)) options
                        else []
                forward =
                    if isFree g (f, 8)
                        then List.map (promotion' (f, r) (f, 8)) options
                        else []
                right =
                    if isOpposite g (f + 1, 8)
                        then List.map (promotion' (f, r) (f + 1, 8)) options
                        else []
            in  left ++ forward ++ right
        else []

potentialWhitePawnMoves : Game -> Position -> List Move
potentialWhitePawnMoves g p =
    let normal    = potentialWhitePawnNormalMoves  g p
        enPassant = potentialWhiteEnPassant        g p
        promotion = potentialWhitePromotions       g p
    in normal ++ enPassant ++ promotion

potentialBlackPawnNormalMoves : Game -> Position -> List Move
potentialBlackPawnNormalMoves g =  \(f, r) ->
    if r == 2
        then []
        else
            let left =
                    if isOpposite g (f - 1, r - 1)
                        then [normal' (f, r) (f - 1, r - 1)]
                        else []
                forward =
                    if isFree g (f, r - 1)
                        then [normal' (f, r) (f, r - 1)]
                        else []
                fastForward =
                    if r == 7 && isFree g (f, 6) && isFree g (f, 5)
                        then [normal' (f, 7) (f, 5)]
                        else []
                right =
                    if isOpposite g (f + 1, r - 1)
                        then [normal' (f, r) (f + 1, r - 1)]
                        else []
            in  left ++ forward ++ fastForward ++ right

potentialBlackEnPassant : Game -> Position -> List Move
potentialBlackEnPassant g = \(f, r) ->
    case g.enPassant of
        Nothing         -> []
        Just (epf, epr) ->
            if r - 1 == epr && abs (f - epf) == 1
                then [enPassant' (f, r) (epf, epr)]
                else []

potentialBlackPromotions : Game -> Position -> List Move
potentialBlackPromotions g = \(f, r) ->
    if r == 2
        then
            let options = [BlackKnight, BlackBishop, BlackRook, BlackQueen]
                left =
                    if isOpposite g (f - 1, 1)
                        then List.map (promotion' (f, r) (f - 1, 1)) options
                        else []
                forward =
                    if isFree g (f, 1)
                        then List.map (promotion' (f, r) (f, 1)) options
                        else []
                right =
                    if isOpposite g (f + 1, 1)
                        then List.map (promotion' (f, r) (f + 1, 1)) options
                        else []
            in  left ++ forward ++ right
        else []

potentialBlackPawnMoves : Game -> Position -> List Move
potentialBlackPawnMoves g p =
    let normal    = potentialBlackPawnNormalMoves  g p
        enPassant = potentialBlackEnPassant        g p
        promotion = potentialBlackPromotions       g p
    in normal ++ enPassant ++ promotion

{-
Generates potential knight moves. It is not tested whether the king is in check
after the move, so some of them might be illegal.
-}
potentialKnightMoves : Game -> Position -> List Move
potentialKnightMoves g = \(f, r) ->
    let toNormal p =
        if isFree g p || isOpposite g p
            then Just <| normal' (f, r) p
            else Nothing
    in List.filterMap toNormal <|
        [ (f - 2, r - 1), (f - 2, r + 1)
        , (f - 1, r - 2), (f - 1, r + 2)
        , (f + 1, r - 2), (f + 1, r + 2)
        , (f + 2, r - 1), (f + 2, r + 1) ]

potentialSlidingMoves : Game -> Position -> (File -> Int -> Int) -> (Rank -> Int -> Int) -> List Move
potentialSlidingMoves g = \(f, r) df dr ->
    let line            = List.map (\i -> (df f i, dr r i)) [1..7]
        (free, blocked) = span (isFree g) line
    in List.map (normal' (f, r)) <|
            free ++ case blocked of
                []      -> []
                (x::xs) -> if isOpposite g x then [x] else []

potentialBishopMoves : Game -> Position -> List Move
potentialBishopMoves g p =
    let slide = potentialSlidingMoves g p
    in     slide (-) (-)
        ++ slide (-) (+)
        ++ slide (+) (-)
        ++ slide (+) (+)

potentialRookMoves : Game -> Position -> List Move
potentialRookMoves g p =
    let slide = potentialSlidingMoves g p
    in     slide (-)    always
        ++ slide always (-)
        ++ slide always (+)
        ++ slide (+)    always

potentialQueenMoves : Game -> Position -> List Move
potentialQueenMoves g p =
    let slide = potentialSlidingMoves g p
    in     slide (-)    (-)
        ++ slide (-)    always
        ++ slide (-)    (+)
        ++ slide always (-)
        ++ slide always (+)
        ++ slide (+)    (-)
        ++ slide (+)    always
        ++ slide (+)    (+)

potentialKingNormalMoves : Game -> Position -> List Move
potentialKingNormalMoves g = \(f, r) ->
    let toNormal p =
        if isFree g p || isOpposite g p
            then Just <| normal' (f, r) p
            else Nothing
    in List.filterMap toNormal <|
        [ (f - 1, r - 1), (f - 1, r    )
        , (f - 1, r + 1), (f    , r - 1)
        , (f    , r + 1), (f + 1, r - 1)
        , (f + 1, r    ), (f + 1, r + 1) ]

potentialKingsideCastling : Game -> List Move
potentialKingsideCastling g =
    if g.player == White
        then if g.castling.whiteKingside && isFree g f1 && isFree g g1 && (not <| isCheck { g | player <- Black }) && (isLegal g <| Normal { from = e1, to = f1})
            then [KingsideCastling { to = g1 }]
            else []
        else if g.castling.blackKingside && isFree g f8 && isFree g g8 && (not <| isCheck { g | player <- White }) && (isLegal g <| Normal { from = e8, to = f8})
            then [KingsideCastling { to = g8 }]
            else []

potentialQueensideCastling : Game -> List Move
potentialQueensideCastling g =
    if g.player == White
        then if g.castling.whiteQueenside && isFree g (Position.d, 1) && isFree g c1 && isFree g b1 && (not <| isCheck { g | player <- Black }) && (isLegal g <| Normal { from = e1, to = d1})
            then [QueensideCastling { to = c1 }]
            else []
        else if g.castling.blackQueenside && isFree g (Position.d, 8) && isFree g c8 && isFree g b8 && (not <| isCheck { g | player <- White }) && (isLegal g <| Normal { from = e8, to = d8})
            then [QueensideCastling { to = c8 }]
            else []

potentialKingMoves : Game -> Position -> List Move
potentialKingMoves g p =
    let normal            = potentialKingNormalMoves   g p
        kingsideCastling  = potentialKingsideCastling  g
        queensideCastling = potentialQueensideCastling g
    in normal ++ kingsideCastling ++ queensideCastling

potentialMoves : Game -> Position -> List Move
potentialMoves g p =
    let (Just piece) = Dict.get p g.board
    in case piece of
        WhitePawn   -> potentialWhitePawnMoves g p
        BlackPawn   -> potentialBlackPawnMoves g p
        WhiteKnight -> potentialKnightMoves    g p
        BlackKnight -> potentialKnightMoves    g p
        WhiteBishop -> potentialBishopMoves    g p
        BlackBishop -> potentialBishopMoves    g p
        WhiteRook   -> potentialRookMoves      g p
        BlackRook   -> potentialRookMoves      g p
        WhiteQueen  -> potentialQueenMoves     g p
        BlackQueen  -> potentialQueenMoves     g p
        WhiteKing   -> potentialKingMoves      g p
        BlackKing   -> potentialKingMoves      g p

allPotentialMoves : Game -> List Move
allPotentialMoves g =
    let isOwnPiece _ piece = piece `belongsTo` g.player
        positions = Dict.keys <| Dict.filter isOwnPiece g.board
    in List.concatMap (potentialMoves g) positions

legalMoves : Game -> Position -> List Move
legalMoves g p = List.filter (isLegal g) <| potentialMoves g p

allLegalMoves : Game -> List Move
allLegalMoves g = List.filter (isLegal g) <| allPotentialMoves g

applyMove : Game -> Move -> Game
applyMove g m =
    case m of
        Normal    { from, to }        -> normal from to g
        EnPassant { from, to }        -> enPassant from to g
        Promotion { from, to, piece } -> promotion from to piece g
        KingsideCastling _            -> kingsideCastling g
        QueensideCastling _           -> queensideCastling g

isLegal : Game -> Move -> Bool
isLegal g m =
    let newGame = applyMove g m
    in not <| isCheck newGame



showNormal : Game -> Move -> String
showNormal g = \(Normal {from, to}) ->
    let (Just piece) = Dict.get from g.board
        delimiter    = case Dict.get to g.board of
            Nothing -> "-"
            Just _  -> "x"
    in Piece.show piece ++ Position.show from ++ delimiter ++ Position.show to



showEnPassant : Game -> Move -> String
showEnPassant g m = ""

showPromotion : Game -> Move -> String
showPromotion g = \(Promotion {from, to, piece}) ->
    Position.show from ++ "-" ++ Position.show to ++ "=" ++ Piece.show piece

show : Game -> Move -> String
show g m =
    case m of
        Normal _            -> showNormal    g m
        EnPassant _         -> showEnPassant g m
        Promotion _         -> showPromotion g m
        KingsideCastling _  -> "O-O"
        QueensideCastling _ -> "O-O-O"
